getCOMErrorCode <-
function(name)
{
 .COMErrors[name]
}


COMError <-
function(code, description = "", className = "COMError")
{
  if(is.character(code))
    code <- getCOMErrorCode(code)

  x <- list(code = as.numeric(code), description = description)
  class(x) <- className
  x
}
			computeCOMNames <-
#
#  Compute the collection of names used in this
#  collection.
#
function(els)
{
  funcs <- els[sapply(els, is.function)]
  n <- as.character(unlist(sapply(funcs, function(x) names(formals(x)))))

  i <- match(".properties", names(els))
  if(!is.na(i))
   propertyNames <- els[[".properties"]]
  else
   propertyNames <- character(0)

  n <- unique(c(names(funcs), n, propertyNames, ""))
  ans <- 1:length(n)
  names(ans) <- n
  ans
}

COMSIDispatchObject <-
#
# Generator for handling IDispatch at the S level.
# 
function(funs, properties = character(0),
	.propertyEnvironment = environment(funs[sapply(funs, is.function)][[1]]))
{
  if(length(properties))
    funs[[".properties"]] <- properties

  .ids <- computeCOMNames(funs)

  hasProperty <-
   function(name) {
     name %in% funs[[".properties"]]
   }

  getProperty = function(name) get(name, envir = .propertyEnvironment)
  setProperty = function(name, val) assign(name, val, envir = .propertyEnvironment)

  invoke <-
   function(id, method, args, namedArgs) {

     if(length(namedArgs)) {
       names(args)[1:length(namedArgs)] <- names(.ids)[namedArgs]
     }
     args <- rev(args)

     name <- names(.ids)[id]

     if(method[2] && !method[1] && !hasProperty(name)) {
       return(COMError("DISP_E_MEMBERNOTFOUND", className = c("COMReturnValue", "COMError")))
     }
	
        # If the name of the method/property is not in the list of functions
        # and the invocation type is for setting or getting a property,
        # then deal with properties.
     if(!(name %in% names(funs)) &  any(method[-1])) {
       if(method[2])
         return(getProperty(name))
       else
         return(setProperty(name, args[[1]]))
     } else if(method[1]) {
         # we are dealing with a method name.
       f <- funs[[name]]
       eval(as.call(c(f, args)), env = globalenv())
     } else {
         # Looking for an element which corresponds to a function but
         # method[1] is not set so can't call it.
       return(COMError("DISP_E_MEMBERNOTFOUND", className = c("COMReturnValue", "COMError")))
     }
   }

  getIDsOfNames <-
      # Lookup the integer identifier for a vector of names.
    function(ids) {
      i <- match(ids, names(.ids))       
      if(any(is.na(i))) {
        stop(paste("Names not found: ", paste(ids[is.na(i)], collapse=", "), " in ", paste(names(.ids), collapse=", " )))
      }

      .ids[i]
    }

 # Adding a third function means a destructor function.
  list(Invoke = invoke, GetIDsOfNames = getIDsOfNames)	
}
			

# setAs("function", "SCOMMethod", function(from) {
#                              x = new("SCOMMethod")
#                              x@Function = from
#                              x
#                            })

	library(methods)

if(FALSE) {	
setClass("SCOMClassElement",
         representation(name="character",
                        help="character"))

# Should classes be identified by name or a class object.
# The latter is preferrable, but 

setClass("SCOMProperty",
          representation("SCOMClassElement",
                         type="character",
                         readOnly="logical"))


setClass("SCOMMethod",
          representation("SCOMClassElement",
                         returnType="character",
                         signature="character",
                         Function = "function"
                         ))

}

setClass("SCOMClass",
          representation(name = "character",
	                 version = "character", # Should this be a real version object.
	                 classId = "Ruuid",
			 progId = "Ruuid",
			 help="character",
                         "VIRTUAL"))

#
setClass("SCOMEnvironmentClass",
	  representation("SCOMClass",
	                 generator = "function"))

setClass("SCOMFunctionClass",
	  representation("SCOMClass",
	                 functions="list",
                         properties = "character"))


setClass("SCOMIDispatch",
	  representation("SCOMEnvironmentClass"))	
.initSCOMClass <-
function(obj, name, help, clsid = NULL)
{
 obj@name = name
 obj@help = help

 if(is.null(clsid)) {
   if(require(Ruuid))
     obj@classId = getuuid()
 }

 invisible(obj)
}


if(FALSE) {	
 # Ignore for now.
SCOMGenericFunctionClass <-
function(functions, name, propertyNames = character(), ..., help = "",
         def = new("SCOMGenericFunctionClass"), where="")
{
 def <- .initSCOMClass(def, name = name, help = help) 
 def@functions = c(normal="rnorm", poisson="rpois", binomial="rbinom")
 def@propertyNames = propertyNames
 def@properties = list(...)
 
 def
}

SCOMEnvironmentClass <-
function(generator, name, help = "", def = new("SCOMEnvironmentClass"), where="")
{
 def <- .initSCOMClass(def, name = name, help = help) 
 def@generator = generator

 def
}
} # FALSE

SCOMFunctionClass <-
function(functions, properties = character(0), name, help = "", def = new("SCOMFunctionClass"), where="")
{
 def <- .initSCOMClass(def, name = name, help = help) 
 def@functions = functions
 def@properties = properties

 def
}

	
SCOMIDispatch <-
function(generator, name, help = "", def = new("SCOMIDispatch"), where="")
{
 def <- .initSCOMClass(def, name = name, help = help) 
 def@generator = generator

 def
}
	

setGeneric("createCOMObject", function(def) standardGeneric("createCOMObject"))

setMethod("createCOMObject", "SCOMEnvironmentClass", function(def) {
             .Call("R_RCOMEnvironmentObject", def, PACKAGE="RDCOMServer")
   })


setMethod("createCOMObject", "SCOMIDispatch", function(def) {
	     funcs <- def@generator()
             obj <- COMSIDispatchObject(funcs)
             .Call("R_RCOMSObject", obj, PACKAGE="RDCOMServer")
   })

setMethod("createCOMObject", "matrix", function(def) {
             obj <- COMSIDispatchObject(matrixCOMHandler(def))
             .Call("R_RCOMSObject", obj, PACKAGE="RDCOMServer")
   })	

# values doesn't really make sense here.
setMethod("createCOMObject", "data.frame", function(def) {
             obj <- COMSIDispatchObject(matrixCOMHandler(def))
             .Call("R_RCOMSObject", obj, PACKAGE="RDCOMServer")
   })	

setMethod("createCOMObject", "SCOMFunctionClass", function(def) {
             obj <- COMSIDispatchObject(def@functions, def@properties)
             .Call("R_RCOMSObject", obj, PACKAGE="RDCOMServer")
   })

setMethod("createCOMObject", "SCOMGenericFunctionClass", function(def) {
             obj <- COMSIDispatchObject(def@functions)
             .Call("R_RCOMSObject", obj, PACKAGE="RDCOMServer")
   })

setMethod("createCOMObject", "ANY", function(def) {
             obj <- genericCOMHandler(def)
             .Call("R_RCOMSObject", obj, PACKAGE="RDCOMServer")
   })



matrixCOMHandler <-
function(m)
{
 f=list(dim = function() dim(m),
 	column = function(j) m[,j, drop = TRUE],
        row = function(i) m[i,, drop = TRUE],
        element = function(i,j) m[i,j],
        nrow = function() nrow(m),
        ncol = function() nrow(m),
        rownames = function() rownames(m),
        colnames = function() colnames(m),
        .properties = character(0)
       )

 if(is.matrix(m))
   f[["values"]] = function() as.vector(m)
  
 f
}	
genericCOMHandler <-
#
# The idea is very simple. We present a COM object
# associated with an arbitrary S object and 
# we respond to property requests by looking 
# for named elements either in the object itself
# or in its attributes.
# 
# We can respond to methods by calling the 
# corresponding S function with the object as
# the first argument.
#
function(obj)	
{
  .nameIDs <- integer(0)

  getIDsOfNames <-  function(ids) {

# cat("Looking for ids for", ids, "\n") 
    idx <- match(ids, names(.nameIDs))

     # Any names that aren't there, just add them to the list of known names.
    if(any(is.na(idx))) {
       n <- ids[is.na(idx)]
       tmp <- seq(from = length(.nameIDs) + 1, length = length(n))
       names(tmp) <- n

       .nameIDs <<- c(.nameIDs, tmp)
    }

    as.integer(.nameIDs[ids])
  }


  hasProperty <- 
     function(name) {
       els <- c(names(obj), names(attributes(obj)))
       return(!is.na(match(name, els)))
     }

  getProperty <-
     function(name) {
       if(!is.null(names(obj))) {
         idx <- match(name, names(obj))  
         if(!is.na(idx))
          return(obj[[name]])
       }

       if(!is.null(names(attributes(obj)))) {
         idx <- match(name, names(attributes(obj)))
         if(!is.na(idx))
          return(attr(obj, name))
       }

       NULL
     }

  invoke <- 
    function(id, method, args, namedArgs) {
if(id == 0) {
#  cat("Id of 0 passed to invoke\n")
  return(COMError("DISP_E_MEMBERNOTFOUND", className = c("COMReturnValue", "COMError")))
}

     name <- names(.nameIDs)[id]

# cat("method name", name, "from", id, "\n")
# cat("method name", hasProperty(name),"\n")		
     if(method[2] && hasProperty(name)) {
       val <- getProperty(name)
       return(val)
     }

     if(method[1]) {
        if(exists(name, mode = "function")) {
#          cat("[Generic] Calling", name, "\n")
   	  if(is.null(args))
	    args <- list(obj)
	  else
	    args <- c(obj, args)
          return(do.call(name, args))
	}
     }

     stop("[Generic COM Object::invoke] Can't handle ", name)
   }

   list(Invoke = invoke, GetIDsOfNames = getIDsOfNames)
}
getCOMClassDefFileName <-
function()
{
  val = Sys.getenv("R_COM_CLASS_FILE")
  if(val != "")
    return(val)

  if("RCOMClassFile" %in% names(options()))
   return(options()[["RCOMClassFile"]])
  
  paste(.find.package("RDCOMServer"), "RCOMClasses.rda", sep = .Platform$file.sep)
}

GetCOMClassDef <-
function(clsid, rda = getCOMClassDefFileName(), verbose = FALSE)	
{

  # Look for an agreed on entries in the registry for this CLSID
  # to see if we should load the class definition object from
  # elsewhere. If we can, then load it and return.
  # Check if a key exists.
  
 if(verbose)	
   cat("Looking for CLSID", clsid, "\n")


 if(require(SWinRegistry)) {

   # patch up the clsid with { } in case we are given it with no { }, i.e. from Ruuid.
  if(substring(clsid, 1,1) != "{") 
   clsid <- paste("{", clsid, "}", sep="")


   # See if there are any entries in the CLSID's folder
   # in the registry indicating how to create an object.

# registryKeyExists(c("CLSID", clsid, "profile"))
  tmp = getRegistryValue(c("CLSID", clsid), "profile", isError = FALSE)
  if(!is.na(tmp) && tmp != "") {
    if(file.exists(tmp))
      source(tmp)
    else {
      e = try(parse(text = tmp))
      if(!inherits(e, "try-error"))
	eval(e, envir = globalenv())
    }
  }

  tmp = getRegistryValue(c("CLSID", clsid), "rda", isError = FALSE)
  if(!is.na(tmp) && tmp != "") {
     env = new.env()
       # Want this to be a URI.
     tmp = getRegistryValue(c("CLSID", clsid), "rda")
     val = try(load(tmp, envir = env)) 
     if(!inherits(val, "try-error")) {
	val = get(val[1], envir = env)
        return(val)
     } else {
      cat("Couldn't load RDA file", tmp, "\n")
     }
  } else {
    tmp = getRegistryValue(c("CLSID", clsid), "Svariable", isError = FALSE)
    if(!is.na(tmp) && tmp != "")  {
       val = get(tmp)
       return(val)
    }
  }
 } else
   cat("Can't load SWinRegistry package\n")

 if(substring(clsid, 1,1) == "{") {
   clsid <- substring(clsid, 2, nchar(clsid)-1)
 }

 defs <- loadCOMClassDefs(rda)

 clsid <- toupper(clsid)
 which <- match(clsid, names(defs))
 if(is.na(which)) {
   stop("No COM class definition for class ID ", clsid, ".  Available UUIDs known to R", paste(names(defs), collapse=", "))
 }

 def <- defs[[which]]

 if(verbose && !is.null(def)) 
   cat("Found definition for CLSID", clsid, " ", def@name, "\n")

 def
}


setClass("SCOMNamedFunctionClass",
	  representation("SCOMClass",
              # can be named character vector to provide aliases for COM view.
              # e.g. functions=c(normal="rnorm", ...)
	                 functions="character",
                         propertyNames="character",
                         properties="list"))

SCOMNamedFunctionClass =
 function(functions, name, ..., propertyNames = character(0),
          help = "", def = new("SCOMNamedFunctionClass"))
{
  def = .initSCOMClass(def, name = name, help = help) 

  def@functions = functions
  if(length(names(def@functions)) == 0)
    names(def@functions) = def@functions
  which = names(def@functions) == ""
  if(any(which))
     names(def@functions)[which] = def@functions[which]

  def@propertyNames = propertyNames
  def@properties = list(...)

  def
}


COMNameManager <-
function(map = integer(0))
{
 if(length(map) && !(is.numeric(map) && !is.null(names(map)))) {
   tmp <- 1:length(map)
   names(tmp) <- as.character(map)
   map <-  tmp
 }


 lookup <-
   function(name) {
 
     if(missing(name))
      return(map)

     if(is.numeric(name)) {
       return(names(map)[name])
     }

     idx <- match(name, names(map))

     if(any(is.na(idx))) {
       n <- name[is.na(idx)]
       orig <- length(map) + 1
       tmp <- seq(from = orig, length = length(n))
       names(tmp) <- n

       map <<- c(map, tmp)        
       idx[is.na(idx)] <- seq(orig,  length = length(n))
     }

cat(name, "=>", idx, "\n")
     as.integer(idx)
   }

  lookup
}



SCOMNamedFunctionDispatch =
#
# Generator for handling IDispatch at the S level.
# 
function(funs, properties, propertyNames, nameMgr = COMNameManager())
{
  isProperty = function(name) {
    name %in% c(names(properties), propertyNames)
  }

  setProperty = function(name, val) {
    if(name %in% names(properties))
      properties[[name]] <<- val
    else if(name %in% propertyNames)
      assign(name, val, envir = globalenv())
  }

  getProperty = function(name) {
    if(name %in% names(properties))
      properties[[name]]
    else if(name %in% propertyNames)
      get(name, envir = globalenv())
  }

  invoke =
   function(id, method, args, namedArgs) {
     if(length(namedArgs)) {
       names(args)[1:length(namedArgs)] = nameMgr(namedArgs)
     }
     args = rev(args)

     methodName = nameMgr(id)

     if(!(methodName %in% names(funs)) &&  any(method[-1])) {
       if(!isProperty(methodName)) 
         return(COMError("DISP_E_MEMBERNOTFOUND", className = c("COMReturnValue", "COMError")))
         
       if(method[2]) {
          getProperty(methodName)
       } else {
          setProperty(methodName, args[[1]])
       }
     } else {
       if(!method[1])
         return(COMError("DISP_E_MEMBERNOTFOUND", className = c("COMReturnValue", "COMError")))
       else
         eval(do.call(funs[methodName], args), envir = globalenv())
     }
  }

  list(Invoke= invoke, GetNamesOfIDs = nameMgr)	
}


setMethod("createCOMObject", "SCOMNamedFunctionClass", 
   function(def) {
     obj = SCOMNamedFunctionDispatch(def@functions, def@properties, def@propertyNames)
    .Call("R_RCOMSObject", obj)
   })
registerCOMClass <-
function(def, clsid = def@classId, registry = TRUE, 
	  rda = getCOMClassDefFileName(), ...)
{

 if(length(rda) == 1)
  rda <- rep(rda, 2)

 defs <- loadCOMClassDefs(rda[1])

 if(is.null(defs))
  defs <- list()

 defs[[toupper(as.character(clsid))]] <- def
 saveCOMClassDefs(rda[2], defs)

 if(registry)
    registerClassID(def@name, clsid, version = def@version, ...)

  TRUE
}

unregisterCOMClass <-
 function(name, clsid = NULL, registry = TRUE,
	  rda = getCOMClassDefFileName())
{
  if(is.null(clsid))
    clsid <- getRegistryValue(c(name, "CLSID"), "")

  if(registry)
    unregisterClassID(name, clsid)

  if(length(rda) == 1)
    rda <- rep(rda, 2)

  defs <- loadCOMClassDefs(rda[1])
  which <- match(id, names(defs))
  if(!is.na(which)) {
     ans <- defs[[which]]
     defs <- defs[-which]
     saveCOMClassDefs(rda[2], defs)
  } else 
     stop("Can't find class definition for CLSID ", id)

  ans   
}

	
saveCOMClassDefs <-
function(fileName, COMDefs)
{
 save(COMDefs, file = fileName)

 TRUE
}

loadCOMClassDefs <-
function(fileName = getCOMClassDefFileName())
{
 if(!file.exists(fileName)) 
   return(NULL)

 library(methods)

 env <- new.env()
 load(fileName, env)
 defs <- get(objects(envir = env)[1], envir=env)

 invisible(defs)
}
registerClassID <-
function(name, clsid = getuuid(), progId = name, appId = NULL, version = "1.0", 
	  dll = system.file("RDCOMEngine.dll", package = "RDCOMServer"),
	  exe = system.file("RDCOM.exe", package = "RDCOMServer"),
	  addClassID = FALSE, useEXE = FALSE,
          check = TRUE,
          rda = NULL, profile = NULL, ...,
          force = FALSE)
{
 if(!require(SWinRegistry)) {
  stop("Cannot register COM class identifier in registry. Install SWinRegistry")
 } 

  clsid <- paste("{", toupper(as.character(clsid)), "}", sep="")

  if(useEXE)  {
   tmp <- gsub("{", "", clsid)
   tmp <- gsub("}", "", tmp)
   dll <- paste(exe, tmp)
   procKey = "LocalServer32"
  } else
    procKey = "InprocServer32"

  dll <- gsub("/", "\\\\", dll)

  if(check) {
    if(isRegistered(name))  {
      id <- lookupCLSIDByName(name)
      if(id != clsid && !force) {
       stop("Mismatch between Class IDs: original registry value ", id, " != ", clsid)
      }
    }
  }

  createRegistryKey("CLSID", clsid)
    setRegistryValue(c("CLSID", clsid), "", value = name)

    if(!is.null(appId)) {
      setRegistryValue(c("CLSID", clsid), "AppID", value=appId)
    }

    createRegistryKey(c("CLSID", clsid), procKey)
    setRegistryValue(c("CLSID", clsid, procKey), "", value=dll)

    if(!is.null(progId)) {
      createRegistryKey(c("CLSID", clsid), "ProgID")
      setRegistryValue(c("CLSID", clsid, "ProgID"),  "", value=progId)
    }

  extraKeys = c(rda = rda, profile = profile, ...)
  for(i in names(extraKeys)) {
    if(!is.null(extraKeys[i])) 
     setRegistryValue(c("CLSID", clsid), i, value=as.character(extraKeys[i]))
  }

# This should be the progId (?)

  createRegistryKey(name)

    createRegistryKey(name, "CLSID")
    setRegistryValue(c(name, "CLSID"), "", value=clsid)

    if(length(version)) {
     createRegistryKey(name, "CurVer")
     setRegistryValue(c(name, "CurVer"), "", value=version)
    }

  names(clsid) <- name
  invisible(clsid)
}

unregisterClassID <-
#
# In the future, if purge = FALSE, then we remove only the
# elements we set in registerClassID and do not just remove
# the entire folders.
#
# progId is only needed if purge = FALSE.
#
function(name, clsid = NULL, progId = name, purge = TRUE)
{
  if(is.null(clsid)) 
    clsid <- getRegistryValue(c(name, "CLSID"), "")

  if(substring(clsid, 1, 1) != "{")
    clsid <- paste("{", toupper(as.character(clsid)), "}", sep="") 

  if(purge) {
     contents <- getRegistry(c("CLSID", clsid))
     deleteRegistryKey(c("CLSID", clsid), asKey = TRUE, recursive = TRUE)
     return(contents)
  } else {
    # stop("Only purge = TRUE is currently supported by unregisterClassID") 
  
      keys = list(createRegistryPath(c("CLSID", clsid, "InprocServer32", ""), isValue = TRUE),
                  createRegistryPath(c("CLSID", clsid, "InprocServer32")),
		  createRegistryPath(c("CLSID", clsid, "LocalServer32", ""), isValue = TRUE),
                  createRegistryPath(c("CLSID", clsid, "LocalServer32")),
		  createRegistryPath(c("CLSID", clsid, "ProgID", ""), isValue = TRUE),
                  createRegistryPath(c("CLSID", clsid, "ProgID")),
                  createRegistryPath(c("CLSID", clsid)),
                  )

      for(i in keys) {
         if(registryKeyExists(i)) {
           if(is(i, "RegistryKeyPath") ||
                 (length(getRegistrySubKeyNames(i)) == 0 && length(getRegistryKeyValues(i)) == 0))
             deleteRegistryKey(i)
         }
      }

    if(FALSE) { # turned off for now.
      keys <- getRegistrySubKeyNames(c("CLSID", clsid))
      for(i in keys)
        deleteRegistryKey(c("CLSID", clsid), i, asKey = TRUE)

      deleteRegistryKey("CLSID", clsid, asKey = TRUE) 

      keys <- getRegistySubKeyNames(name)
      for(i in keys)
       deleteRegistryKey(name, i, asKey = TRUE)

      deleteRegistryKey(name, clsid, asKey = TRUE) 
    }
  }

  TRUE
}

lookupCLSIDByName <-
function(name)
{
  library(SWinRegistry)
  getRegistryValue(c(name, "CLSID"), "")
}

isRegistered <-
function(name)
{
 library(SWinRegistry)	
 if(registryKeyExists(name))
   name <- lookupCLSIDByName(name)

 registryKeyExists(c("CLSID", name))
}

					

.First.lib <-
function(lib, pkg)
{
 library(methods)
 library(Ruuid) # , verbose = FALSE)

 library.dynam("RDCOMServer", pkg, lib)
 assign(".COMErrors", .Call("R_createCOMErrorCodes", PACKAGE = pkg), "package:RDCOMServer")
}
